unit NkVImage;

interface

uses Windows, ImgFilter;

type
  TMaxBitmapInfo = packed record
    bmiHeader: TBitmapInfoHeader;
    bmiColors: array[0..255+3] of TRGBQuad;
  end;

  TNkvImage = class(TObject)
  private
    FBitmapInfo: TMaxBitmapInfo;
    FBits: PChar;
    procedure Scale(var DestBmi: TMaxBitmapInfo; DestBits: Pointer;
      DestWidth, DestHeight, ViewLeft, ViewTop, ViewWidth, ViewHeight: Longint;
      Filter: Boolean);
  public
    Loaded: Boolean;
    Width: Longint;
    Height: Longint;
    BitsSize: Cardinal;
    Bpp: Cardinal;
    OrgSize: Cardinal;
    FormatName: string;
    Comment: string;
    FilterName: string;
    FilterFileName: string;
    procedure ReadMem(Buf: PChar; BufSize: Cardinal; const Filter: TImgFilter);
    procedure Draw(hDC: HDC; Left, Top, AWidth, AHeight: Longint;
      DestLeft, DestTop, DestWidth, DestHeight: Longint;
      Grid: Boolean; ScaleFilter: Boolean);
    procedure Write(const FileName: string; AWidth, AHeight: Longint;
      ScaleFilter: Boolean);
    procedure GetPixel(X, Y: Longint; var R, G, B: Integer);
  end;

implementation

uses NTypes, NLib, ImgScaler, SpiFilter;

procedure GetImageFormat(Buf: PChar; var MimeType: string; var Name: string);
begin
  MimeType := '';
  Name := '';
  if Copy(Buf, 1, 3) = 'GIF' then
  begin
    MimeType := 'image/gif';
    Name := Copy(Buf, 1, 6);
  end
  else if Copy(Buf, 1, 2) = #$FF#$D8 then
  begin
    Inc(Buf, 6);
    if Copy(Buf, 1, 4) = 'JFIF' then
    begin
      MimeType := 'image/jpeg';
      Name := 'JPEG (JFIF)';
    end
    else if Copy(Buf, 1, 4) = 'Exif' then
    begin
      MimeType := 'image/jpeg';
      Name := 'Exif';
    end;
  end
  else if Copy(Buf, 1, 8) = #$89'PNG'#$0D#$0A#$1A#$0A then
  begin
    MimeType := 'image/png';
    Name := 'PNG';
    case Buf[28] of
      #0: Name := Name + ' - non-interlaced';
      #1: Name := Name + ' - interlaced';
    end;
    if Ord(Buf[25]) and 4 <> 0 then
      Name := Name + ' (with alpha)';
  end;
end;

procedure TNkVImage.ReadMem(Buf: PChar; BufSize: Cardinal; const Filter: TImgFilter);
var
  size, i: Integer;
  bmi: TMaxBitmapInfo;
  tmpbits: PChar;
  imginfo: TImgInfo;
  mimetype: string;
  fname: array[0..MAX_PATH] of Char;
  w, h, c: Cardinal;
begin
  Filter.GetSize(Buf, BufSize, w, h, c);

  size := (w * c + 3) and not 3 * h;
  // If Delphi's memory manager returned splitted memory block,
  // SetDIBitsToDevice() may fail. So use GlobalAlloc().
  tmpbits := Pointer(GlobalAlloc(0, size));
  if tmpbits = nil then
    raise EOSError.Create;

  try
    Filter.Decode(Buf, BufSize, tmpbits, size, imginfo);
  except
    GlobalFree(HGLOBAL(tmpbits));
    raise;
  end;

  // Load success

  FillChar(bmi.bmiHeader, SizeOf(bmi.bmiHeader), 0);
  with bmi.bmiHeader do
  begin
    biSize := SizeOf(bmi.bmiHeader);
    biWidth := w;
    biHeight := h;
    biPlanes := 1;
    biBitCount := c * 8;
    biCompression := BI_RGB;
    biSizeImage := size;
  end;

  if c = 1 then // gray scale
    for i := 0 to 255 do
    begin
      bmi.bmiColors[i].rgbBlue := i;
      bmi.bmiColors[i].rgbGreen := i;
      bmi.bmiColors[i].rgbRed := i;
      bmi.bmiColors[i].rgbReserved := 0;
    end;

  if FBits <> nil then
    GlobalFree(HGLOBAL(FBits));
  FBitmapInfo := bmi;
  FBits := tmpbits;
  BitsSize := size;
  Width := w;
  Height := h;
  Bpp := imginfo.bpp;
  if imginfo.FormatName = '' then
    GetImageFormat(Buf, mimetype, FormatName)
  else
    FormatName := imginfo.FormatName;
  //Comment := ConvLineBreak(imginfo.Comment, #13#10);
  Comment := imginfo.Comment; // XXX
  FilterName := Filter.Name;
  if Filter is TSpiFilter then
  begin
    GetModuleFileName((Filter as TSpiFilter).Handle, fname, SIzeOf(fname));
    FilterFileName := fname;
  end else
    FilterFileName := '';
  OrgSize := BufSize;
  Loaded := True;
end;

procedure TNkVImage.Scale(var DestBmi: TMaxBitmapInfo; DestBits: Pointer;
  DestWidth, DestHeight, ViewLeft, ViewTop, ViewWidth, ViewHeight: Longint;
  Filter: Boolean);
begin
  ViewTop := DestHeight - ViewTop - ViewHeight; // for bottom-up DIB
  with FBitmapInfo.bmiHeader do
    if Filter then
    begin
      if (DestWidth > biWidth) or (biWidth div DestWidth <= 2) then
        Bilinear(FBits, DestBits, biWidth, biHeight, biBitCount,
          DestWidth, DestHeight, ViewLeft, ViewTop, ViewWidth, ViewHeight)
      else
        Average(FBits, DestBits, biWidth, biHeight, biBitCount,
          DestWidth, DestHeight, ViewLeft, ViewTop, ViewWidth, ViewHeight);
    end else
      Nearest(FBits, DestBits, biWidth, biHeight, biBitCount,
        DestWidth, DestHeight, ViewLeft, ViewTop, ViewWidth, ViewHeight);

  DestBmi := FBitmapInfo;
  DestBmi.bmiHeader.biWidth := ViewWidth;
  DestBmi.bmiHeader.biHeight := ViewHeight;
end;

procedure TNkVImage.Draw(hDC: HDC; Left, Top, AWidth, AHeight: Longint;
  DestLeft, DestTop, DestWidth, DestHeight: Longint; Grid: Boolean;
  ScaleFilter: Boolean);
var
  bmi: TMaxBitmapInfo;
  tmpbits: Pointer;
  x, y, pitch, a, b, c: Integer;
  z: Double;
  tmp: PChar;
begin
  if (DestWidth = Width) and (DestHeight = Height) and not Grid then
    SetDIBitsToDevice(hDC, Left, Top, DestWidth, DestHeight, DestLeft, 0,
      DestTop, DestHeight, FBits, PBitmapInfo(@FBitmapInfo)^, DIB_RGB_COLORS)
  else if (DestHeight <> 0) and (DestWidth <> 0) then
  begin
    pitch := CalcPitch(AWidth, FBitmapInfo.bmiHeader.biBitCount);

    // If Delphi's memory manager returned splitted memory block,
    // SetDIBitsToDevice() may fail. So use GlobalAlloc().
    tmpbits := Pointer(GlobalAlloc(GMEM_FIXED, pitch * AHeight));
    Scale(bmi, tmpbits, DestWidth, DestHeight, DestLeft, DestTop,
      AWidth, AHeight, ScaleFilter);

    if Grid then
    begin
      c := 0;
      z := Width / DestWidth;
      b := Trunc(z * (DestLeft-1+0.5));
      for x := 0 to AWidth - 1 do
      begin
        a := Trunc(z * (DestLeft+x+0.5));
        if b <> a then
        begin
          b := a;
          if z > 0.25 then
          begin
            if b and 15 = 0 then
              c := 0
            else if b and 7 = 0 then
              c := 128
            else
              continue;
          end else
            if b and 7 = 0 then
              c := 0
            else
              c := 128;
          if FBitmapInfo.bmiHeader.biBitCount = 24 then
          begin
            tmp := @PChar(tmpbits)[x*3];
            for y := 0 to AHeight - 1 do
            begin
              tmp[0] := Char(c);
              tmp[1] := Char(c);
              tmp[2] := Char(c);
              Inc(tmp, pitch);
            end;
          end else
            for y := 0 to AHeight - 1 do
              PChar(tmpbits)[y*pitch+x] := Char(c);
        end;
      end;

      z := Height / DestHeight;
      b := Trunc(z * (DestTop-1+0.5));
      for y := 0 to AHeight - 1 do
      begin
        a := Trunc(z * (DestTop+y+0.5));
        if b <> a then
        begin
          b := a;
          if z > 0.25 then
          begin
            if b and 15 = 0 then
              c := 0
            else if b and 7 = 0 then
              c := 128
            else
              continue;
          end else
            if b and 7 = 0 then
              c := 0
            else
              c := 128;
          FillChar(PChar(tmpbits)[(AHeight-y-1)*pitch], pitch, c);
        end;
      end;
    end;

    SetDIBitsToDevice(hDC, Left, Top, AWidth, AHeight, 0, 0, 0, AHeight,
      tmpbits, PBitmapInfo(@bmi)^, DIB_RGB_COLORS);
    GlobalFree(HGLOBAL(tmpbits));
  end;
end;

procedure TNkVImage.Write(const FileName: string; AWidth, AHeight: Longint;
  ScaleFilter: Boolean);
var
  bfh: TBitmapFileHeader;
  bmi: TMaxBitmapInfo;
  bmisize, bitssize: Cardinal;
  tmpbits: Pointer;
begin
  bitssize := Longint(CalcPitch(AWidth, FBitmapInfo.bmiHeader.biBitCount))
    * AHeight;
  if (AWidth = Width) and (AHeight = Height) then
  begin
    bmi := FBitmapInfo;
    tmpbits := FBits;
  end else
  begin
    GetMem(tmpbits, bitssize);
    Scale(bmi, tmpbits, AWidth, AHeight, 0, 0, AWidth, AHeight, ScaleFilter);
  end;
  try
    try
      with TFile.Create(FileName, [fmWrite, fmTruncate]) do
        try
          with bmi do
          begin
            bmisize := SizeOf(bmiHeader);
            if bmiHeader.biBitCount <= 8 then
              Inc(bmisize, (1 shl bmiHeader.biBitCount) * SizeOf(TRGBQUAD));
          end;

          FillChar(bfh, SizeOf(bfh), 0);
          bfh.bfType := Byte('B') + Byte('M') shl 8;
          bfh.bfSize := SizeOf(bfh) + bmisize + bitssize;
          bfh.bfOffBits := SizeOf(bfh) + bmisize;
          Write(bfh, SizeOf(bfh));
          Write(bmi, bmisize);
          Write(tmpbits^, bitssize);
        finally
          Free;
        end;
    except
      on E: EIOError do raise;
    end;
  finally
    if tmpbits <> FBits then
      FreeMem(tmpbits);
  end;
end;

procedure TNkVImage.GetPixel(X, Y: Longint; var R, G, B: Integer);
var
  p: PChar;
  bpp: Integer;
begin
  bpp := FBitmapInfo.bmiHeader.biBitCount;
  p := @FBits[(Height-Y-1)*Longint(CalcPitch(Width, bpp)) + X*(bpp div 8)];
  if bpp = 24 then
  begin
    B := Ord(p[0]);
    G := Ord(p[1]);
    R := Ord(p[2]);
  end
  else begin
    B := Ord(p[0]);
    G := B;
    R := B;
  end;
end;

end.
